import matplotlib
from matplotlib import pyplot as plt
import json
import operator
from typing import List

from binary_tree import BASEDIR

matplotlib.use('tkagg')  # for use plt.show().


def main(exclude_datalist_names: List[str] = None) -> None:
    with open(BASEDIR / 'performance.json', 'r') as json_data_file:
        data = json.load(json_data_file)

    class_names = sorted(data.keys())
    tests_names = sorted(data[class_names[0]], key=operator.itemgetter(1))

    plot_numbers = len(tests_names)

    fig, axs = plt.subplots(1, plot_numbers, figsize=(9, 3), sharey=True)

    for plot_number, tests_name in enumerate(tests_names):
        curr_axs = axs[plot_number]
        for class_name in class_names:
            plt_data = sorted(data[class_name][tests_name].items())
            if exclude_datalist_names:
                plt_data = [i for i in plt_data if not any(i[0] == exc for exc in exclude_datalist_names)]

            x_data_items, y_data_items = zip(*plt_data)
            curr_axs.plot(x_data_items, y_data_items, label=class_name)

            curr_axs.set_title(tests_name)
            curr_axs.legend()
            plt.setp(curr_axs.xaxis.get_majorticklabels(), rotation=90)

    plt.show()


if __name__ == '__main__':
    # main(exclude_datalist_names=['datalist_1000000'])
    main()



"""
Examples of binary tree, made in the style of OOP

It's just example of binary sort functions, based or not on his own tree.
"""

from abc import ABCMeta, abstractmethod
from collections.abc import Iterator
from numbers import Integral
import bisect


class BaseNodeClass(metaclass=ABCMeta):
    """
    Abstract Binary Tree Class
    All other based on this interface
    """

    @abstractmethod
    def add_node(self, *args) -> None:
        """
        add items to data container
        """
        raise NotImplementedError

    @abstractmethod
    def tree_data(self) -> Iterator:
        """
        return sorted items of inner data
        """
        raise NotImplementedError


class BisectNodeClass(BaseNodeClass):
    """
    Based on bisect stdlib, without inner tree
    """
    def __init__(self):
        self.data = []

    # @profile
    def add_node(self, data: Integral) -> None:
        bisect.insort(self.data, data)

    # @profile
    def tree_data(self) -> Iterator:
        yield from self.data


class SingleNodeClass(BaseNodeClass):
    """
    All operations and storage takes place inside one class
    """

    def __init__(self, data: Integral = None):
        self.left = None
        self.right = None
        self.data = data

    # @profile
    def add_node(self, data: Integral) -> None:
        if self.data is not None:
            if data < self.data:
                if self.left is None:
                    self.left = SingleNodeClass(data)
                else:
                    self.left.add_node(data)
            elif data > self.data:
                if self.right is None:
                    self.right = SingleNodeClass(data)
                else:
                    self.right.add_node(data)
        else:
            self.data = data

    # @profile
    def tree_data(self) -> Iterator:
        if self.left:
            yield from self.left.tree_data()
        yield self.data
        if self.right:
            yield from self.right.tree_data()


class TwoNodeClass(BaseNodeClass):
    """
    Use inner class Node for storage and TwoNodeClass for for everything else
    based on https://gist.github.com/samidhtalsania/6659380
    """

    class Node:
        """
        Inner class of item, which contain info about neightbours
        """
        def __init__(self, key: Integral):
            self.key = key
            self.left = None
            self.right = None
            self.parent = None

    def __init__(self):
        self.root = None

    # @profile
    def add_node(self, key: Integral, _node: Node = None) -> None:

        if _node is None:
            _node = self.root

        if self.root is None:
            self.root = self.Node(key)

        else:
            if key <= _node.key:
                if _node.left is None:
                    _node.left = self.Node(key)
                    _node.left.parent = _node
                    return
                else:
                    return self.add_node(key, _node=_node.left)
            else:
                if _node.right is None:
                    _node.right = self.Node(key)
                    _node.right.parent = _node
                    return
                else:
                    return self.add_node(key, _node=_node.right)

    # @profile
    def tree_data(self, node: Node = None) -> Iterator:
        if node is None:
            node = self.root

        stack = []
        while stack or node:
            if node is not None:
                stack.append(node)
                node = node.left
            else:
                node = stack.pop()
                yield node.key
                node = node.right


if __name__ == '__main__':
    import time

    from data import datalist_100

    def profile():
        nodes = [SingleNodeClass, TwoNodeClass, BisectNodeClass]
        for cls_node in nodes:
            node = cls_node()
            for i in datalist_100:
                node.add_node(i)
            list(node.tree_data())
            del node
            time.sleep(0.3)

    profile()



from collections import defaultdict
import timeit
import json
from contextlib import contextmanager
import time
from typing import List, Callable, Dict, Union
from numbers import Integral

from knapsack_problem import BASEDIR
from knapsack_problem.data import create_dynamic_knapsacks
from knapsack_problem.ref_func import Knapsack, Item
from knapsack_problem.funcs import (
    knapsack_1_standard_solution,
    knapsack_2_solution,
    knapsack_3_solution,
    knapsack_4_bruteforce_solution,
    knapsack_5_dynamic_solution,
    knapsack_6_recursive_dynamic_solution,
    knapsack_greedy_solution
)

RETRY_NUM = 1000
TOO_LONG = 60  # in seconds
SETUP_IMPORT_TEMPLATE = '''
from __main__ import {func_name} as knapsack_solution, {datalist_name} as raw_datalist
datalist = raw_datalist['{datalist_key}']
'''
RUNNING_TEMPLATE = "raw_result = knapsack_solution(*datalist['input'])"
RESULT_TO_LIST_TEMPLATE = RUNNING_TEMPLATE + '\nraw_result = list(raw_result)'
SMALL_STAT_NAME = 'small_perf'
BIG_STAT_NAME = 'big_perf'


def mean(numbers: List[Integral]) -> int:
    return sum(numbers) / len(numbers) / RETRY_NUM


@contextmanager
def time_time(msg: str) -> None:
    start = time.monotonic()
    yield start
    print('{} done: '.format(msg), time.monotonic() - start)


perf_funcs_without_bruteforce = [
    knapsack_1_standard_solution,
    knapsack_2_solution,
    knapsack_3_solution,
    knapsack_5_dynamic_solution,
    knapsack_6_recursive_dynamic_solution,
    knapsack_greedy_solution
]
perf_funcs_with_bruteforce = perf_funcs_without_bruteforce[:] + [knapsack_4_bruteforce_solution]

result = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))


def small_perf() -> None:
    common_perf_part(funcs=perf_funcs_with_bruteforce, datalists=perf_small_datalists, stat_name=SMALL_STAT_NAME)


def big_perf() -> None:
    common_perf_part(funcs=perf_funcs_without_bruteforce, datalists=perf_datalists, stat_name=BIG_STAT_NAME)


def common_perf_part(*,
                     funcs: List[Callable],
                     datalists: Dict[str, Dict[str, Union[Knapsack, List[Item]]]],
                     stat_name: str) -> None:
    datalists_name = next(k for k, v in globals().items() if v == datalists)
    for func in funcs:
        func_name = func.__name__
        print('\n' + '*' * 10, func_name, '*' * 10)
        for datalist_name, _ in list(datalists.items()):
            print('\n' + datalist_name)

            with time_time(stat_name) as start_time:
                result[stat_name][func_name][datalist_name] = mean(
                    timeit.repeat(
                        RESULT_TO_LIST_TEMPLATE,
                        setup=SETUP_IMPORT_TEMPLATE.format(
                            func_name=func_name,
                            datalist_name=datalists_name,
                            datalist_key=datalist_name),
                        number=RETRY_NUM
                    )
                )
            if time.monotonic() - start_time > TOO_LONG:
                break


def main() -> None:
    small_perf()
    big_perf()
    with open(BASEDIR / 'performance.json', 'w') as outfile:
        json.dump(result, outfile, indent=4)
        print('Done')


if __name__ == '__main__':
    perf_small_datalists = create_dynamic_knapsacks(start=3, end=30, step=3)
    perf_datalists = create_dynamic_knapsacks(start=30, end=450, step=30)
    main()



"""
Abstract helper module for check correct other solutions in tests
Also define "Item" and "Knapsack" abstract types
"""
import sys
from functools import lru_cache
from collections import namedtuple
from typing import Tuple

sys.setrecursionlimit(10_000)


class Item(namedtuple('Item', 'name value weight')):
    """
    Atomic thing with his own characteristics

    .. py:attribute:: name:

        Name of item

    .. py:attribute:: value:

        Cost of item

    .. py:attribute:: weight:

        Weight of item
    """


class Knapsack(namedtuple('Knapsack', 'items weight_limit')):
    """
    Knapsack characteristics

    .. py:attribute:: items:

        one of Item above

    .. py:attribute:: weight_limit:

        limit of Knapsack for items
    """


def knapsack_standard_solution(items: Tuple[Item], weight_limit: int) -> Item:
    """
    Standard solution for knapsack problem in Python
    https://codereview.stackexchange.com/a/20581
    """

    @lru_cache(maxsize=None)
    def best_value(i: int, j: int):
        if j < 0:
            return float('-inf')
        if i == 0:
            return 0
        value, weight = items[i - 1].value, items[i - 1].weight
        res = max(best_value(i - 1, j),
                  best_value(i - 1, j - weight) + value)
        return res

    for k in reversed(range(len(items))):
        if best_value(k + 1, weight_limit) != best_value(k, weight_limit):
            yield items[k]
            weight_limit -= items[k].weight



import math
import sys
from typing import List, Dict, Tuple, Union, Iterator
Num = Union[int, float]

sys.setrecursionlimit(11000)


def create_data_decreasing_depth(
        data: Union[List, Iterator],
        length: int,
        max_depth: int,
        _current_depth: int = None,
        _result: List = None
) -> List:
    """
    creates data in depth on decreasing.

    Examples:

    >>> data = create_data_decreasing_depth(list(range(1, 11)), length=5, max_depth=3)
    >>> assert data == [[[1, 2, 3, 4, 5], 6, 7, 8, 9, 10]]
    >>> data = create_data_decreasing_depth(list(range(1, 11)), length=2, max_depth=3)
    >>> assert data == [[[1, 2], 3, 4], 5, 6], [[7, 8,] 9, 10]]
    """
    _result = _result or []
    _current_depth = _current_depth or max_depth
    data = iter(data)
    if _current_depth - 1:
        _result.append(create_data_decreasing_depth(
            data=data,
            length=length,
            max_depth=max_depth,
            _current_depth=_current_depth - 1,
            _result=_result))
    try:
        _current_length = length
        while _current_length:
            item = next(data)
            _result.append(item)
            _current_length -= 1

        if max_depth == _current_depth:
            _result += create_data_decreasing_depth(
                data=data,
                length=length,
                max_depth=max_depth)
        return _result

    except StopIteration:
        return _result


def create_data_increasing_depth(
        data: Union[List, Iterator],
        length: int,
        max_depth: int,
        _current_depth: int = None,
        _result: List = None
) -> List:
    """
    creates data in depth to increase.

    Examples:

    >>> data = create_data_increasing_depth(list(range(1, 11)), length=5, max_depth=3)
    >>> assert data == [1, 2, 3, 4, 5, [6, 7, 8, 9, 10]]
    >>> data = create_data_increasing_depth(list(range(1, 11)), length=2, max_depth=3)
    >>> assert data == [1, 2, [3, 4, [5, 6]]], 7, 8, [9, 10]]
    """
    _result = _result or []
    _current_depth = _current_depth or max_depth
    data = iter(data)
    try:
        _current_length = length
        while _current_length:
            item = next(data)
            _result.append(item)
            _current_length -= 1

    except StopIteration:
        return _result

    if _current_depth - 1:
        tmp_res = create_data_increasing_depth(
            data=data,
            length=length,
            max_depth=max_depth,
            _current_depth=_current_depth - 1)
        if tmp_res:
            _result.append(tmp_res)

    if max_depth == _current_depth:
        tmp_res = create_data_increasing_depth(
            data=data,
            length=length,
            max_depth=max_depth)
        if tmp_res:
            _result += tmp_res
    return _result


def generate_data() -> List[Tuple[str, Dict[str, Union[range, Num]]]]:
    """
    Generated collections of Data by pattern

    {amount_item}_amount_{length}_length_{max_depth}_max_depth

    where:

    .. py:attribute:: amount_item:

        len of flatten elements

    .. py:attribute:: length:

        len of elements at the same level of nesting

    .. py:attribute:: max_depth:

        highest possible level of nesting
    """

    data = []
    amount_of_elements = [10 ** i for i in range(5)]
    data_template = '{amount_item}_amount_{length}_length_{max_depth}_max_depth'

    # amount_item doesn't need to be [1]
    for amount_item in amount_of_elements[1:]:
        for max_depth in amount_of_elements:
            # for exclude flatten list after generate data by create_data_increasing_depth
            if amount_item > max_depth:
                # generate four types of length
                for length in range(0, max_depth + 1, math.ceil(max_depth / 4)):
                    # min length must be 1
                    length = length or 1

                    data_name = data_template.format(
                        amount_item=amount_item,
                        length=length,
                        max_depth=max_depth
                    )

                    data_value = {
                        'data': range(amount_item),
                        'length': length,
                        'max_depth': max_depth
                    }

                    data.append((data_name, data_value))

                    # for not to produce more than 1 flat entity
                    if max_depth == 1:
                        break

    # this order is convenient for me
    data = sorted(data, key=lambda x: [x[1]['data'][-1], x[1]['max_depth'], x[1]['length']])
    return data


def get_data_name_order() -> List[str]:
    return [i[0] for i in generate_data()]


if __name__ == '__main__':
    for data in generate_data():
        print(data[0])



"""
Generate test data for test performance
"""
from typing import List
from random import shuffle


def generate_data(num: int) -> List[int]:
    """
    generate data numbers in random order
    result -> "datalist_*num*"
    """
    data = list(range(num))
    shuffle(data)
    return data


datalist_100 = generate_data(100)
datalist_1000 = generate_data(1000)
datalist_10000 = generate_data(10000)
datalist_100000 = generate_data(100000)
datalist_1000000 = generate_data(1000000)


import matplotlib
from matplotlib import pyplot as plt
import json
from typing import List

from unpacking_flatten_lists import BASEDIR
from unpacking_flatten_lists.data import get_data_name_order
from unpacking_flatten_lists.performance import INCREMENT_MODE_NAME, DECREMENT_MODE_NAME


AVERAGE = 'average'
matplotlib.use('tkagg')  # for use plt.show().


def main(exclude_funcs: List[str] = None) -> None:
    with open(BASEDIR / 'performance.json', 'r') as json_data_file:
        data = json.load(json_data_file)

    if exclude_funcs is None:
        func_names = sorted(data.keys())
    else:
        func_names = sorted(set(data.keys()) - set(exclude_funcs))

    modes = sorted(data[func_names[0]].keys()) + [AVERAGE]
    data_names_with_order = get_data_name_order()

    plot_numbers = len(modes)
    fig, axs = plt.subplots(1, plot_numbers, figsize=(9, 3), sharey=True)

    for plot_number, mode in enumerate(modes):
        curr_axs = axs[plot_number]
        for func_name in func_names:
            try:
                plt_data = sorted(data[func_name][mode].items(), key=lambda x: data_names_with_order.index(x[0]))
            except KeyError as err:  # for 'average'
                if mode != AVERAGE:
                    raise KeyError from err
                inc_data, dec_data = data[func_name].get(INCREMENT_MODE_NAME), data[func_name].get(DECREMENT_MODE_NAME)
                raw_plt_data = {k: (inc_data.get(k) + dec_data.get(k)) / 2 for k in set(inc_data) & set(dec_data)}
                plt_data = sorted(raw_plt_data.items(), key=lambda x: data_names_with_order.index(x[0]))

            x_data_items, y_data_items = zip(*plt_data)
            curr_axs.plot(x_data_items, y_data_items, label=func_name)

            curr_axs.set_title(mode)
            curr_axs.legend()
            plt.setp(curr_axs.xaxis.get_majorticklabels(), rotation=90)

    plt.show()


if __name__ == '__main__':
    main()
    main(exclude_funcs=['tishka_flatten', 'outer_flatten_2', 'recursive_flatten_iterator'])



import timeit
import json
import time
from collections import defaultdict
from typing import List, Callable
from numbers import Integral
from contextlib import contextmanager

from unpacking_flatten_lists import BASEDIR
from unpacking_flatten_lists.data import (
    generate_data,
    create_data_decreasing_depth,
    create_data_increasing_depth)
from unpacking_flatten_lists.funcs import (
    outer_flatten_1,
    outer_flatten_2,
    niccolum_flatten,
    tishka_flatten,
    zart_flatten,
    recursive_flatten_iterator,
    tishka_flatten_with_stack,
    recursive_flatten_generator)

RETRY_NUM = 10
TOO_LONG = 60 * 5 // 100  # in seconds
INCREMENT_MODE_NAME = 'increase'
DECREMENT_MODE_NAME = 'decrease'
SETUP_IMPORT_TEMPLATE = '''
from typing import Iterator
import json
from __main__ import {func_name} as flatten
data = json.loads("{data}")
'''
RUNNING_TEMPLATE = '''
result = flatten(data)
if isinstance(result, Iterator):
    result = list(result)
'''


funcs = [
    outer_flatten_1,
    outer_flatten_2,
    niccolum_flatten,
    tishka_flatten,
    zart_flatten,
    recursive_flatten_iterator,
    tishka_flatten_with_stack,
    recursive_flatten_generator
]

result = defaultdict(lambda: defaultdict(lambda: defaultdict(Integral)))


def mean(numbers: List[Integral]) -> int:
    return sum(numbers) / len(numbers) / RETRY_NUM


@contextmanager
def time_time(msg: str) -> None:
    start = time.monotonic()
    yield start
    print('{} done: '.format(msg), time.monotonic() - start)


def increase_part():
    print('*' * 10, 'Increase', '*' * 10)
    common_part(data_create_func=create_data_increasing_depth, mode=INCREMENT_MODE_NAME)


def decrease_part():
    print('*' * 10, 'Decrease', '*' * 10)
    common_part(data_create_func=create_data_decreasing_depth, mode=DECREMENT_MODE_NAME)


def common_part(*, data_create_func: Callable, mode: str):
    for func in funcs:
        func_name = func.__name__

        print('\n', func_name, '\n')

        for data_example in generate_data():
            data = data_create_func(**data_example[1])
            data = json.dumps(data)  # crutch because timeit has s_push: parser stack overflow for list with 100 deep
            data_struct_name = data_example[0]

            with time_time(data_struct_name) as start_time:
                result[func_name][mode][data_struct_name] = mean(
                    timeit.repeat(
                        RUNNING_TEMPLATE,
                        setup=SETUP_IMPORT_TEMPLATE.format(
                            func_name=func_name,
                            data=data),
                        number=RETRY_NUM
                    )
                )
            if time.monotonic() - start_time > TOO_LONG:
                break


def main():
    increase_part()
    decrease_part()

    print('Done testing. Writes...')
    with open(BASEDIR / 'performance.json', 'w') as outfile:
        json.dump(result, outfile, indent=4)
        print('Done')


if __name__ == '__main__':
    main()


from collections import defaultdict
import timeit
import json
from contextlib import contextmanager
import time
from typing import List
from numbers import Integral

from binary_tree import BASEDIR
from binary_tree.funcs import SingleNodeClass, TwoNodeClass, BisectNodeClass
from binary_tree.data import datalist_100, datalist_1000, datalist_10000, datalist_100000, datalist_1000000

RETRY_NUM = 100
TOO_LONG = 60*2


def mean(numbers: List[Integral]) -> int:
    return sum(numbers) / len(numbers) / RETRY_NUM


@contextmanager
def time_time(msg: str) -> None:
    start = time.monotonic()
    yield start
    print('{} done: '.format(msg), time.monotonic() - start)


perf_classes = [SingleNodeClass, TwoNodeClass, BisectNodeClass]
perf_datalists = [datalist_100, datalist_1000, datalist_10000, datalist_100000, datalist_1000000]

result = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))

fill_nodes_command = """
instance = Node()
for data in datalist:
    instance.add_node(data)
"""

print_node_list_command = """
result = list(instance.tree_data())
"""

setup_import_template = 'from __main__ import {cls_name} as Node, {datalist_name} as datalist'


def main():
    for cls in perf_classes:
        cls_name = cls.__name__
        print(cls_name)

        for datalist in perf_datalists:
            datalist_name = next(k for k, v in globals().items() if v == datalist)
            print(datalist_name)

            with time_time('add_node'):
                result[cls_name]['add_node'][datalist_name] = mean(
                    timeit.repeat(
                        fill_nodes_command,
                        setup=setup_import_template.format(cls_name=cls_name, datalist_name=datalist_name),
                        number=RETRY_NUM
                    )
                )

            with time_time('tree_data'):
                result[cls_name]['tree_data'][datalist_name] = mean(
                    timeit.repeat(
                        print_node_list_command,
                        setup=(setup_import_template + fill_nodes_command).format(
                            cls_name=cls_name, datalist_name=datalist_name
                        ),
                        number=RETRY_NUM
                    )
                )

            with time_time('full') as start_time:
                result[cls_name]['full'][datalist_name] = mean(
                    timeit.repeat(
                        (fill_nodes_command + print_node_list_command),
                        setup=setup_import_template.format(
                            cls_name=cls_name, datalist_name=datalist_name
                        ),
                        number=RETRY_NUM
                    )
                )
            if time.monotonic() - start_time > TOO_LONG:
                break

    with open(BASEDIR / 'performance.json', 'w') as outfile:
        json.dump(result, outfile, indent=4)
        print('Done')


if __name__ == '__main__':
    main()



import matplotlib
from matplotlib import pyplot as plt
import json

from knapsack_problem import BASEDIR
from knapsack_problem.performance import SMALL_STAT_NAME, BIG_STAT_NAME

matplotlib.use('tkagg')  # for use plt.show().


def main(*, stat_type) -> None:
    with open(BASEDIR / 'performance.json', 'r') as json_data_file:
        data = json.load(json_data_file)[stat_type]

    func_names = sorted(data.keys())

    fig, axs = plt.subplots()

    for func_name in func_names:
        # sorted by int in datalist name for correct sort by ascending items count in a knapsack
        plt_data = sorted(data[func_name].items(), key=lambda x: int(x[0].split('_')[1]))
        x_data_items, y_data_items = zip(*plt_data)
        axs.plot(x_data_items, y_data_items, label=func_name)

        axs.set_title(stat_type)
        axs.legend()
        plt.setp(axs.xaxis.get_majorticklabels(), rotation=90)

    plt.show()


if __name__ == '__main__':
    # main(stat_type=SMALL_STAT_NAME)
    main(stat_type=BIG_STAT_NAME)




import os
import sys

from setuptools import setup

if sys.version_info < (3, 6):
    raise RuntimeError("Intrst_algrms 3.x requires Python 3.6+")


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()


setup(
    name="Intrst_algrms",
    author="Nikolai Vidov",
    author_email="lastsal@mail.ru",
    version=read('VERSION'),
    description="https://intrst-algrms.readthedocs.io/en/latest/",
    license="MIT",
    keywords="example documentation algorithms",
    url="https://pypi.org/project/Intrst-algrms/",
    packages=['.', 'binary_tree', 'knapsack_problem', 'unpacking_flatten_lists'],
    package_data={
        # If any package contains files, include them:
        '': ['*.txt', '*.rst', '*.md', '*.mdn', '*.png', '*.lprof', '*.json', '*.dat'],
    },
    python_requires='>=3.6',
    install_requires=[
        'Cython',
        'iteration-utilities',
        'more-itertools',
        'memory-profiler',
        # 'line-profiler', # install line-profiler manually
        'matplotlib',
    ],
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
)
